from promptdesignerdataset.dataset import PromptDesignerDataset, LoadJsonData, SaveJsonData
from copy import deepcopy
from itertools import chain
from collections import OrderedDict


class Pipeline:
    INPUT_TYPE_ENTIRE = 'file-type-entire'
    INPUT_TYPE_LIST = 'file-type-list'

    def __init__(self):
        self.data = {'file': None, # this file path
                    'input-files' : list(),

                    'output-script-modules': list(),
                    'output-item-filter-modules': list(),
                    # 'input-text-filename': None,
                    # 'input-type': None,

                    'apply-script-output': list(),
                    'script-name': None,
                    'pipeline-name': None,
                    'pipeline': list(),

                    'gpt3-settings': {},
                    'steps': 0,

                     'prompts': list()}

    def AddPrompt(self,
                  prompt_name,
                  prompt_template):
        #  add a prompt template
        self.data['prompts'].append(tuple([prompt_name,
                                           prompt_template]))

    def GetPromptNames(self):
        #  return prompts names
        prompt_names = [name for name, template in self.data['prompts']]

        return prompt_names

    def GetPrompt(self,
                  prompt_name):
        #  return a prompt template

        prompt_item = list(filter(lambda x: x[0] == prompt_name, self.data['prompts']))[0]
        prompt_template = prompt_item[1]

        return prompt_template

    def RemoveItem(self,
                   item_name):

        for n_, item in enumerate(self.GetItems()):
            if item.Name() == item_name:
                self.data['pipeline'].pop(n_)


    def __len__(self):
        return len(self.data['pipeline'])

    def AddOutputScriptModule(self, module_path):
        self.data['output-script-modules'].append(module_path)

    def AddOutputFilter(self,
                        filter_name):
        self.data['apply-script-output'].append(filter_name)

    def ClearOutputFilters(self):
        self.data['apply-script-output'] = list()

    def GetOutputFilters(self):
        return self.data['apply-script-output']

    def GetOutputScriptModules(self):
        return self.data['output-script-modules']

    def AddItemOutputScriptModule(self, module_path):
        self.data['output-item-filter-modules'].append(module_path)

    def GetItemOutputScriptModules(self):
        return self.data['output-item-filter-modules']

    def SetFile(self,
                file_):
        self.data['file'] = file_

    def GetFile(self):
        return self.data['file']

    def GetGPT3Settings(self):
        return self.data['gpt3-settings']

    def AddGPT3Settings(self,
                        settings):
        self.data['gpt3-settings'] = settings

    def SetName(self, name):
        self.data['pipeline-name'] = name

    def GetName(self):
        return self.data['pipeline-name']

    def Clear(self):
        self.data['input-files'].clear()
        self.data['output-script-modules'].clear()
        self.data['output-item-filter-modules'].clear()
        self.data['apply-script-output'].clear()
        self.data['pipeline'].clear()
        self.data['prompts'].clear()

        self.reset_step_counter()

    def _increment_step(self):
        self.data['steps'] += 1

    def reset_step_counter(self):
        self.data['steps'] = 0

    def total_steps(self):
        return self.data['steps']

    def AddItem(self,
                item,
                autoset_step=True):
        print(type(item))
        #  autoset set automatically set the step,
        # use false in load to manually set the step
        self._increment_step()
        if autoset_step:
            item.SetStep(self.total_steps())
        self.data['pipeline'].append(item)

    def PrepareDataToSave(self):
        #  return a dict of the pipeline data
        data_to_save = deepcopy(self.data)
        data_to_save['pipeline'] = [item.GetItemDict() for item in self.GetItems()]
        return data_to_save

    def SavePipeline(self, filename):
        data_to_save = self.PrepareDataToSave()
        data_to_save['file'] = filename

        return SaveJsonData(data_to_save, filename)

    def AddInputFile(self, inputfile_dict):
        # 'file_': None,
        # 'input-type': None,
        # 'content'
        # 'mnemonic-name
        self.data['input-files'].append(inputfile_dict)

    def GetInputFilesMnemonicNames(self):
        return [item['mnemonic-name'] for item in self.data['input-files']]

    def GetItem(self,
                item_name):
        for item in self.data['pipeline']:
            if item.Name() == item_name:
                return item
        raise AttributeError('Item not present in the pipeline')

    def GetInputFileContent(self,
                            mnemonic_name):
        for inp in self.data['input-files']:
            if inp['mnemonic-name'] == mnemonic_name:
                return inp['content']

    def GetInputFilesItems(self):
        return self.data['input-files']

    def GetItems(self):
        items = list()

        unsorted = [item for item in self.data['pipeline']]
        return sorted(unsorted, key=lambda x: x.GetStep())

    def LoadPipelilne(self, filename):
        data = LoadJsonData(filename)
        if data:
            # promptdesignerdataset = deepcopy(promptdesignerdataset)
            try:

                self.SetName(data['pipeline-name'])
                #  add input files
                for fileitem in data['input-files']:
                    self.AddInputFile(fileitem)

                self.AddGPT3Settings(data['gpt3-settings'])
                self.SetFile(data['file'])

                for mod in data['output-script-modules']:
                    self.AddOutputScriptModule(mod)

                for mod in data['output-item-filter-modules']:
                    self.AddItemOutputScriptModule(mod)
                #  add output scripts to list box
                for filter in data['apply-script-output']:
                    self.AddOutputFilter(filter)

                for item in data['pipeline']:
                    # input_item = {'apply-item-filters': item['input-item']['apply-item-filters'],
                    #               'input-bindings': item['input-item']['input-bindings'],
                    #               }
                    pitem = PipelineItem()
                    pitem.SetData(item)
                    # pitem.SetData(name=item['name'],
                    #               prompt=item['prompt'],
                    #               input_item=item['input_item'])
                    # pitem.SetStep(item['step'])
                    self.AddItem(deepcopy(pitem))

                for prompt_name, prompt_template in data['prompts']:
                    self.AddPrompt(prompt_name, prompt_template)
            except KeyError:
                print('FILE NOT VALID.')

    def ExportResults(self):
        # return a dict of results
        results = OrderedDict()
        for item in self.GetItems():
            item_data = {'step': item.GetStep(),
                         'results-raw': item.data['results-raw'],
                         'results-filtered': item.data['results-filtered']} #  the last one! }
            results[item.Name()] = item_data
        return results

    def ClearResults(self):
        for item in self.GetItems():
            item.ClearResults()


class PipelineItem:
    """
    name = self.pipeline_item_name.get(),
    input_item = item_input,
    apply_filter_to_output = self.chk_apply_filter_to_pipeline_item_output.get(),
    filter_on_output_name = self.pipeline_output_filter_name.get())

        item_input = {'setting-type': self.experiment_setting_range_type.get(),
                      'apply-filter-script': self.chk_apply_filter_to_pipeline_item_output.get(),
                      'filter-script-name': self.pipeline_item_output_filter_name.get(),
                      'input-bindings': list()
                      }
        item_input['input-bindings'].append({'prompt-key': child[0],
                                                 'key-value': child[1],
                                                 'type': child[2]})

    """

    def __init__(self):
        self.data = {'input_item': {
                                    'apply-item-filters': list(),
                                    'input-bindings': list(),
                                    },
                     'results-raw': list(),
                     'results-filtered': list()}
    def SetData(self,
                data):
        self.data = data


    def SetInput(self,
                 input_item):
        self.data['input_item'] = input_item
    # def SetData(self,
    #              name,
    #              prompt,
    #              input_item,
    #              ):
    #     self.promptdesignerdataset['name'] = name
    #     self.promptdesignerdataset['prompt'] = prompt
    #     self.promptdesignerdataset['input_item'] = input_item

    def SetPrompt(self,
                  prompt):
        self.data['prompt'] = prompt

    def GetPrompt(self):
        return self.data['prompt']

    def SetStep(self, step_n):
        self.data['step'] = step_n

    def GetStep(self):
        return self.data['step']

    def GetItemFilterNames(self):
        """
         Generator of filter to apply to the item output
        :return:
        """
        # ['input_item']{'apply-item-filters': [i for i in self.lst_item_filters.get(0, 'end')],
        #                      # self.chk_apply_filter_to_pipeline_item_output.get(),
        #                      'filter-script-name': self.pipeline_item_output_filter_name.get(),
        #                      'input-bindings': list()
        #                      }
        for filter in self.data['input_item']['apply-item-filters']:
            yield filter

    def GetPromptName(self):
        return self.data['prompt']

    def AddResultsRaw(self,
                      result,
                      key_value_pair):
        #  results is the result of GPT
        #  key_value_pair is the dict of k,v pairs - prompt key, value
        self.data['results-raw'].append({'result': result,
                                         'key-value-pair': key_value_pair})
        # self.promptdesignerdataset['results-raw'] = {'result': result,
        #                             'key-value-pair': key_value_pair}


    def AddResultsFiltered(self,
                           results,
                           # key_value_pair,
                           filter_name):
        # self.promptdesignerdataset['results-filtered'].extend([{'result': results,
        self.data['results-filtered'].append([{'result': results,
        #                                       'key-value-pair': key_value_pair,
                                              'filter-name': filter_name}])


    def GetResultsContent(self):
        # return the results raw if no filter is applied.
        results = list()
        if self.data['results-filtered']:
            for item in self.data['results-filtered'][-1]:
                results.append(item['result'])
            #  flatten the list
            results = list(chain(*results))

        else:
            for item in self.data['results-raw']:
                results.append(item['result'])
            # if len(results) > 1:
            #     results = list(chain(*results))
        return results
    #
    # def GetResults(self):
    #     # return the results raw if no filter is applied.
    #
    #     #  always return a list and a list of dict
    #
    #     results = list()
    #     params = list()
    #
    #     if self.promptdesignerdataset['results-filtered']:
    #         return self.promptdesignerdataset['results-filtered'][-1]
    #         for item in self.promptdesignerdataset['results-filtered'][-1]:
    #             results.append(item.promptdesignerdataset['result'])
    #             params.append()
    #     else:
    #         print('here')
    #
    #         for item in self.promptdesignerdataset['results-raw']:
    #             results.append(item['result'])
    #     return results, params

    def ClearResults(self):
        self.data['results-raw'].clear()
        self.data['results-filtered'].clear()

    def GetPromptParameters(self):
        parameters = dict()
        for row in self.data['input_item']['input-bindings']:
            parameters[row['prompt-key']] = row['key-value']

        return parameters

    def GetItemDict(self):
        return self.data

    def GetTreeItem(self):
        return  (self.data['name'],
                 self.data['prompt'],
                 '; '.join(['({}= {})'.format(item[list(item.keys())[0]], item[list(item.keys())[1]])
                            for item in self.data['input_item']['input-bindings']]).strip(),
                 self.data['input_item']['apply-item-filters'])

    def Name(self):
        return self.data['name']
    def SetName(self, name):
        self.data['name'] = name



import sys
import codecs
from copy import deepcopy
import tkinter as tk
import tkinter.ttk as ttk
import importlib
import importlib.machinery
import importlib.util
import os
import openai
from tkinter import messagebox
from tkinter import filedialog
from tkinter import simpledialog
from tkinter import PhotoImage
from pathlib import Path
from promptdesignerdataset.dataset import PromptDesignerDataset, LoadJsonData, SaveJsonData
from PIL import Image, ImageDraw, ImageTk, ImageFont
from itertools import combinations, product
from pipelinemaker.pipeline import Pipeline, PipelineItem
import pprint

from pipelinemaker.experimentmakerGUI import ExperimentGUI

class Experiment(ExperimentGUI):
    def __init__(self,
                 parent):
        self.parent = parent
        super(Experiment, self).__init__(self.parent)
        # super().__init__(self.parent)

    def NewPipeline(self, *event):
        self.ClearPipeline()

    def _pipeline_name_changed(self, *event):
        self.pipeline.SetName(self.pipeline_name.get())

    def LoadPromptDesignerDataset(self, *event):
        filename = Path(filedialog.askopenfilename(initialdir=os.getcwd(),
                                                   title="Please select a file",
                                                   filetypes=(('all files', '*.*'),
                                                              ('Prompt designer file', '*.answer-data'),
                                                              ),
                                                   defaultextension='.answer-data'
                                                   )
                        )
        if filename.name:
            experimental_base_dataset = PromptDesignerDataset()
            experimental_base_dataset.LoadData(str(filename.absolute()))
            for prompt_name in experimental_base_dataset.GetPromptNames():
                prompt_ = experimental_base_dataset.GetPrompt(prompt_name)
                self.pipeline.AddPrompt(prompt_name=prompt_name,
                                        prompt_template=prompt_)

            self._fill_prompts_combobox()

    def LunchExperimentalPipelineSelectedItem(self, *event):
        if not self.pipeline_tree.selection():
            return
        #  retrieve the item from pipeline tree
        item = self.pipeline_tree.selection()[0]
        values = self.pipeline_tree.item(item)['values']
        item_name = values[0]
        #  retrieve item from pipeline object
        item = self.pipeline.GetItem(item_name)
        # lunch experiment only for that item
        self.LunchItemExperiment(item)

    def LunchItemExperiment(self, item):
        def GetContent(v):
            if v in self.pipeline.GetInputFilesMnemonicNames():
                return self.pipeline.GetInputFileContent(v)
            else:
                prev = self.pipeline.GetItem(v)
                return prev.GetResultsContent()
        ############

        if not self._force_rerun_entire_pipeline.get():
            #  check if a result is present. if so, skip this item
            if item.GetResultsContent():
                return

        prompt_name = item.GetPromptName()
        if prompt_name == self.DO_NOTHING:
            #  raw results are the input of
            param = item.GetPromptParameters()
            v = list(param.values())[0]
            content__ = GetContent(v)
            if type(content__) == str:
                content__ = [content__]
            for content_item in content__:
                item.AddResultsRaw(content_item,
                                   param)
        else:
            prompt = self.pipeline.GetPrompt(prompt_name)
            prompt_parameters = item.GetPromptParameters()
            #  parameters substitution
            for k in list(prompt_parameters.keys()):
                v = prompt_parameters[k]
                prompt_parameters[k] = GetContent(v)
            # compose and collect all the prompt
            #  create all the possible prompts
            prompts_to_ask = list()

            # for each non list parameter, I cast it in a list
            for k, v in prompt_parameters.items():
                if type(v) != list:
                    prompt_parameters[k] = [v]
            keys, values = zip(*prompt_parameters.items())
            permutations_dicts = [dict(zip(keys, v)) for v in product(*values)]
            print('combinations:', len(permutations_dicts), permutations_dicts)
            # print('unique combinations: ', len(set(permutations_dicts)))
            #  create prompts
            for combination in permutations_dicts:
                prompt_to_ask = {'composed-prompt': prompt.format(**combination),
                                 'key-value-pairs': combination}
                prompts_to_ask.append(prompt_to_ask)
            #  ask prompts and collect results
            results = list()
            for prompt_to_ask in prompts_to_ask:
                # print('fix results - remove DEV mode!')
                result = AskGPT3(prompt_to_ask['composed-prompt'],
                                 self.gpt3settings)
                # results.append(result)
                item.AddResultsRaw(result,
                                   prompt_to_ask['key-value-pairs'])

        for filter_name, filter_type in item.GetItemFilterNames():
            # print(filter_name, filter_type)
            if filter_type == self.ITEM_OUTPUT_FILTER_TYPE_RESULTS:
                _res_item = item.GetResultsContent()
                results_filtered = self.imported_item_output_filtes[filter_name](_res_item)
            else:
                #  apply an ad-hoc filter on the item.
                #  this means that the filter might consider multiple items
                #  So, I pass the entire pipeline data
                data = self.pipeline.PrepareDataToSave()
                results_filtered = self.imported_item_output_filtes[filter_name](data)

            item.AddResultsFiltered(results_filtered,
                                    filter_name)

    def LunchExperimentalPipeline(self, *event):
        #  lunch experiments

                #  it is of type list of len 1
                # content = content[0]
        if self._force_rerun_entire_pipeline.get():
            #  avoid adding multiple times the same results
            self.pipeline.ClearResults()

        for item in self.pipeline.GetItems():
            self.LunchItemExperiment(item)

        #  show results
        for item in self.pipeline.GetItems():
            print(item.Name(), item.GetStep())
            print(item.GetResultsContent())
        #  end of experiment
        #  apply outputscript if present
        if self.pipeline.GetOutputFilters():
            self.LunchOutputScript()
        messagebox.showinfo('','Experiment Completed.')

    def LunchOutputScript(self, *event):
        # global_results = list()
        results_pipeline = self.pipeline.ExportResults()
        for filter_name in self.pipeline.GetOutputFilters():
            results = self.imported_output_filters[filter_name](results_pipeline)
            print(filter_name)
            print(results)
        messagebox.showinfo('', 'output script applied')

    def SaveAsPipeline(self, *event):
        filename = Path(filedialog.asksaveasfilename(initialdir=os.getcwd(),
                                                     title="Please select a file",
                                                     filetypes=(('Pipeline file', '*.pipeline'),),
                                                     defaultextension='.pipeline',
                                                     )
                        )
        if filename.name:
            return str(filename.absolute())

    def _show_gpt3_settings(self):
        def close():
            app.destroy()
        app = tk.Toplevel(self.parent)
        app.title(self.selected_setting_filename.get())
        frm = tk.Frame(app)
        frm.pack(side='top', fill='both', expand=True)
        t = ttk.Treeview(frm, columns=(['key', 'value']))
        t.heading('#0', text='#')
        t.heading('#1', text='Key')
        t.heading('#2', text='Value')
        t.column('#0',
                anchor='w',
                width=30,
                minwidth=20,
                stretch=tk.NO)
        t.column('#1', anchor='w', stretch=tk.YES)
        t.column('#2', anchor='w', stretch=tk.YES)
        t.pack(side='left',
               fill='both',
               expand=True)
        ys = ttk.Scrollbar(frm,
                           orient='vertical',
                           command=t.yview)
        t['yscrollcommand'] = ys.set
        ys.pack(side='right', fill='y')
        tk.Button(app,
                  text='Close',
                  anchor='center',
                  command=close).pack(side='top',
                                      fill='x')
        # insert promptdesignerdataset
        for n_, item in enumerate(self.gpt3settings.items()):
            k, v = item

            if type(v) == list:
                v = ' '.join(v).strip()
            t.insert('', n_, text=str(n_), values=(k,v))
            #
            # row = '{key:<25}: {value:<25}'.format(key=k,
            #                                       value=v)
            # t.insert(n_, row)
            #
            # index = '{}.0'.format(n_+1)
            # t.insert(index,
            #          row)
        app.mainloop()

    def InputFilePreview(self):
        #  if no file selected, exit the method
        if not self.input_file_filename:
            return

        def close():
            app.destroy()
        app = tk.Toplevel(self.parent)
        app.title(str(Path(self.input_file_filename).name))
        frm = tk.Frame(app)
        frm.pack(side='top', fill='both', expand=True)
        t = tk.StringVar()
        tk.Label(frm,
                 textvariable=t,
                 anchor='w',
                 justify='left',
                 background='blue',
                 foreground='white',
                 ).pack(side='left',
                        fill='both',
                        expand=True)
        tk.Button(app,
                  text='Close',
                  anchor='center',
                  command=close).pack(side='top',
                                      fill='x')
        # insert promptdesignerdataset
        content = self._load_inputfile_content(self.input_file_filename)
        if type(content) == str:
            t.set(content)
        else:
            t.set('\n'.join(content).strip())
        app.mainloop()

    def _show_prompt(self):
        def close():
            app.destroy()
        app = tk.Toplevel(self.parent)
        app.title(self.selected_prompt.get())
        frm = tk.Frame(app)
        frm.pack(side='top', fill='both', expand=True)
        t = tk.Text(frm,
                    height=23,
                    width=180,
                    wrap='word', )
        t.pack(side='left',
               fill='both',
               expand=True)
        ys = ttk.Scrollbar(frm,
                           orient='vertical',
                           command=t.yview)
        t['yscrollcommand'] = ys.set
        ys.pack(side='right', fill='y')
        tk.Button(app,
                  text='Close',
                  anchor='center',
                  command=close).pack(side='top',
                                      fill='x')
        # insert promptdesignerdataset
        if self.selected_prompt.get().strip():
            t.insert('1.0', self.pipeline.GetPrompt(self.selected_prompt.get()))
            app.mainloop()

    def LoadItemOutputModuleGUI(self, *event):
        outputscript_file = Path(filedialog.askopenfilename(initialdir=os.getcwd(),
                                                            title="Please select a file",
                                                            filetypes=(('all files', '*.*'),
                                                                          ('Python file', '*.py'),
                                                                          ),
                                                            defaultextension='.py',
                                                            )
                                 )
        if outputscript_file.name:
            self.LoadItemOutputModule(str(outputscript_file.absolute()))
            #  add ref to pipeline object
            self.pipeline.AddItemOutputScriptModule(str(outputscript_file.absolute()))

    def LoadItemOutputModule(self,
                             scriptfile):
        try:
            scriptfile = Path(scriptfile)
            # get name
            mnemonic_outputscript_name = scriptfile.name
            # Import mymodule
            loader = importlib.machinery.SourceFileLoader(mnemonic_outputscript_name,
                                                          str(scriptfile.absolute()))
            spec = importlib.util.spec_from_loader(mnemonic_outputscript_name, loader)
            importedscript = importlib.util.module_from_spec(spec)
            loader.exec_module(importedscript)
            #  filter out class elements that has no a Parse Method
            namespaces = [i for i in dir(importedscript) if not i.startswith('__')]
            # imported_filters = dict()
            for namespace in namespaces:
                if type(importedscript.__dict__[namespace]) == type:
                    # it is a class
                    namespace_object = importedscript.__dict__[namespace]().Parse
                else:
                    # it is a function
                    namespace_object = importedscript.__dict__[namespace]
                # if the imported function/module pass the test it is added
                try:
                    namespace_object('test string')
                    namespace_object(['test list 1', 'test list 2'])
                    self.imported_item_output_filtes [namespace] = namespace_object
                    self._add_imported_item_output_filters_cmb(namespace)
                except TypeError:
                    continue
                except NotImplementedError:
                    continue
        except:
            messagebox.showerror('',
                                 'The module {} is corrupted or missing. please re-load it.'.format(scriptfile))

    def LoadOutputFilterModuleGUI(self, *event):
        outputscript_file = Path(filedialog.askopenfilename(initialdir=os.getcwd(),
                                                            title="Please select a file",
                                                            filetypes=(('all files', '*.*'),
                                                                       ('Python file', '*.py'),
                                                                       ),
                                                            defaultextension='.py',
                                                            )
                                 )
        if outputscript_file.name:
            self.LoadOutputFilterModule(str(outputscript_file.absolute()))
            #  add ref to pipeline object
            self.pipeline.AddOutputScriptModule(str(outputscript_file.absolute()))

    def LoadOutputFilterModule(self,
                                  filename):
        try:
            # get name
            outputscript_file = Path(filename)
            mnemonic_outputscript_name = outputscript_file.name
            # Import mymodule
            loader = importlib.machinery.SourceFileLoader(mnemonic_outputscript_name, str(outputscript_file.absolute()))
            spec = importlib.util.spec_from_loader(mnemonic_outputscript_name, loader)
            importedscript = importlib.util.module_from_spec(spec)
            loader.exec_module(importedscript)
            #  filter out class elements that has no a Parse Method
            namespaces = [i for i in dir(importedscript) if not i.startswith('__')]
            # imported_filters = dict()
            for namespace in namespaces:
                if type(importedscript.__dict__[namespace]) == type:
                    # it is a class
                    try:
                        namespace_object = importedscript.__dict__[namespace]().Parse
                    except AttributeError:
                        continue
                    except TypeError:
                        continue
                else:
                    continue
                    # No function allowed
                    # # it is a function
                    # namespace_object = importedscript.__dict__[namespace]
                # if the imported function/module pass the test it is added
                try:
                    # namespace_object('test string')
                    self.imported_output_filters[namespace] = namespace_object
                    self._add_imported_output_filters_cmb(namespace)

                except TypeError:
                    continue
                except NotImplementedError:
                    continue
        except:
            messagebox.showerror('',
                                 'The module {} is corrupted or missing. please re-load it.'.format(filename))

    def ExportResultsGUI(self):
        filename = Path(filedialog.asksaveasfilename(initialdir=os.getcwd(),
                                                     title="Please select a file",
                                                     filetypes=(('JSON file', '*.json'),),
                                                     defaultextension='.json',
                                                     )
                        )
        if filename.name:
            exported_data = self.pipeline.ExportResults()
            if SaveJsonData(exported_data, str(filename.absolute())):
                messagebox.showinfo('',
                                    'Results Exported.')

    def ShowPipelineItem(self):
        def close():
            app.destroy()

        if not self.pipeline_tree.selection():
            return
        #  retrieve the item from pipeline tree
        item = self.pipeline_tree.selection()[0]
        values = self.pipeline_tree.item(item)['values']
        item_name = values[0]
        #  retrieve item from pipeline object
        item = self.pipeline.GetItem(item_name)

        item_val_dict = item.GetItemDict()

        #  create GUI
        app = tk.Toplevel(self.parent)
        app.title('Pipeline Item: {}'.format(item_val_dict['name']))
        app.geometry("350x150")
        frm = tk.Frame(app)
        frm.pack(side='top', fill='both', expand=True)
        tree = ttk.Treeview(frm, columns=('val'))
        tree.pack(fill='both', expand=True)
        tree.heading('#0', text='text')
        tree.heading('#1', text='value')
        # tree.heading('#2', text='input type')
        # Specify attributes of the columns (We want to stretch it!)
        tree.column('#0', minwidth=50, width=150, anchor='w', stretch=tk.NO)
        tree.column('#1', width=250, anchor='w', stretch=tk.YES)

        tree.insert('',
                      'end',
                      text='Step',
                      values=(str(item.GetStep())))
        val = '"{}"'.format(str(item_val_dict['prompt']))
        tree.insert('',
                      'end',
                      text='prompt name',
                      values=(val))
        print(item_val_dict['prompt'])
        val = ('')
        tree.insert('',
                      'end',
                      iid='input_item',
                      text='input items',
                      values=val)
        tree.insert('input_item',
                      'end',
                      iid='input-bindings',
                      text='input-bindings',
                      values=val)
        tree.insert('input_item',
                      'end',
                      iid='apply-item-filters',
                      text='apply-item-filters',
                      values=val)

        for n_f, filter in enumerate(item_val_dict['input_item']['apply-item-filters']):
            str_filter = '"{} - {}"'.format(*filter)
            tree.insert('apply-item-filters',
                          'end',
                          text='item filter {}'.format(n_f),
                          values=(str_filter)
                        )


        for n_b, binding in enumerate(item_val_dict['input_item']['input-bindings']):
            tree.insert('input-bindings',
                          'end',
                          text='bindings {}'.format(n_b),
                          values=(str(binding)))

    def ShowResults(self):
        #  collect Exproted promptdesignerdataset
        exported_data = self.pipeline.ExportResults()
        #  show promptdesignerdataset
        pp = pprint.PrettyPrinter()
        exported_data_str = pp.pformat(exported_data)
        # print(exported_data_str)

        def close():
            app.destroy()
        app = tk.Toplevel(self.parent)
        app.title('Results')
        frm = tk.Frame(app)
        frm.pack(side='top', fill='both', expand=True)
        t = tk.Listbox(frm)
        t.pack(side='left', fill='both', expand=True)

        for line in exported_data_str.split('\n'):
            t.insert('end', line)
        ys = ttk.Scrollbar(frm,
                           orient='vertical',
                           command=t.yview)
        t['yscrollcommand'] = ys.set
        ys.pack(side='right', fill='y')
        tk.Button(app,
                  text='Close',
                  anchor='center',
                  command=close).pack(side='top',
                                      fill='x')
        app.mainloop()

    ### THIS SHOULD BE MOVED IN prompt_designer_dataset
    def get_prompt_parameters(self,
                              prompt_name: str):

        prompt = self.pipeline.GetPrompt(prompt_name)

        parameter_name = ''
        #  store parameters key name
        parameter_names = list()
        #  store invalid key name
        rejected_names = list()

        index = 0
        begin_index = 0
        look_for_begin = True
        # prompt = "is {x} greater than {y} or it is {something else}"
        while True:
            if look_for_begin:
                found_ind = prompt.find('{', index)
                if found_ind > 0:
                    begin_index = found_ind + 1
                    # move index
                    index = found_ind + 1
                    look_for_begin = False
                else:
                    break
            else:
                found_ind = prompt.find('}', index)
                if found_ind > 0:
                    # record name
                    parameter_name = prompt[begin_index:found_ind]
                    #  check the key name has not space at begin or end
                    if parameter_name == parameter_name.strip():
                        parameter_names.append(parameter_name)
                    else:
                        rejected_names.append(parameter_name)
                    # move index
                    index = found_ind + 1
                    look_for_begin = True
                else:
                    break

        return list(set(parameter_names))

    def _cmb_prompt_changed(self, *event):
        if self.selected_prompt.get() == self.DO_NOTHING:
            self.cmb_input_prompt_key_name['values'] = ['input']
        else:
            keys = self.get_prompt_parameters(self.selected_prompt.get())
            self.cmb_input_prompt_key_name['values'] = keys
        self._clear_input_key_pairs()

            #  restore parameter keys
        # self._cmb_prompt_changed()
    def ClearInputKeyValuePairs(self, *event):
        self._clear_input_key_pairs()
        self._cmb_prompt_changed()

    def ClearPipeline(self, internal_callback=False):
        #  internal callback suppress warning
        if not internal_callback:
            answer = messagebox.askyesno('', 'Are you sure you want to delete the pipeline?')
            if answer == tk.NO:
                return
        #  clear pipeline
        self.pipeline.Clear()
        self.ClearInputKeyValuePairs()
        self._clear_input_values()

        #  clear pipeline tree
        for item in self.pipeline_tree.get_children():
            self.pipeline_tree.delete(item)

    def RemovePipelineItem(self, event):
        if not event.widget.selection():
            return
        if messagebox.askyesno('',
                               'remove item?') == tk.YES:
            item = event.widget.selection()
            values = self.pipeline_tree.item(item)['values']
            item_name = values[0]
            self.pipeline.RemoveItem(item_name=item_name)
            self.pipeline_tree.delete(item)
            print('item ', item, 'removed')
            #  remove from combobox
            cmb_vals = list(self.cmb_input_prompt_input_value['values'])
            cmb_vals.remove(item_name)
            self.cmb_input_prompt_input_value['values'] = cmb_vals

    def AddInputFile(self):
        file_ = self.input_file_filename
        if not file_:
            messagebox.showerror('',
                                 'You must select a file.')
            return
        filename = str(Path(file_).name)
        mnemonic_name = self.input_file_mnemonic_name.get()
        if not mnemonic_name:
            messagebox.showerror('',
                                 'You must set a mnemonic name for this file')
            return
        inputtype = self.input_type.get()
        content = self._load_inputfile_content(file_)

        inputfileitem = {'file': file_,
                         'input-type': inputtype,
                         'content': content,
                         'mnemonic-name': mnemonic_name}

        self.pipeline.AddInputFile(inputfileitem)
        self._add_inputfile_into_filetreelist(filename, mnemonic_name, inputtype)


    def ClearInputFiles(self):
        #  clean tree view widget
        for item in self.input_files_tree.get_children():
            self.input_files_tree.delete(item)
        #  clear combobox widget
        self.cmb_input_prompt_input_value['values'] = []
        self.cmb_input_prompt_input_value.set('')

    def LoadPipeline(self):
        filename = Path(filedialog.askopenfilename(initialdir=os.getcwd(),
                                                   title="Please select a file",
                                                   filetypes=(('all files', '*.*'),
                                                              ('Pipeline file', '*.pipeline'),
                                                              ),
                                                   defaultextension='.pipeline',
                                                   )
                        )
        if filename.name:

            self.pipeline.Clear()
            #  clear input file
            self.ClearInputFiles()
            #  clear previous pipeline
            self.ClearPipeline(internal_callback=True)
            #  clear output script list
            self._clear_output_script()

            #  load pipeline
            self.pipeline.LoadPipelilne(str(filename.absolute()))
            self.output_pipeline_filename.set(self.pipeline.GetFile())

            #  load promptdesignerdataset into GUI
            self.pipeline_name.set(self.pipeline.GetName())

            for fileitem in self.pipeline.GetInputFilesItems():
                self._add_inputfile_into_filetreelist(filename_name=str(Path(fileitem['file']).name),
                                                      mnemonic_name=fileitem['mnemonic-name'],
                                                      inputtype=fileitem['input-type'],)

            self.gpt3settings = self.pipeline.GetGPT3Settings()
            try:
                openai.api_key = self.gpt3settings["api-key"]
            except:
                print('No settings found! no api-key set.')

            #  load pipeline items
            for item in self.pipeline.GetItems():
                #  add item to pipeline tree view
                self.pipeline_tree.insert('',
                                          'end',
                                          text=str(item.GetStep()),
                                          values=item.GetTreeItem())
                #  add item name to input list
                self._add_input_type_item(item.Name())

            self.pipeline_item_name.set('')
            #  fix combo  experiment item
            self.ClearInputKeyValuePairs()

            #  load modules
            for mod in self.pipeline.GetItemOutputScriptModules():
                self.LoadItemOutputModule(mod)

            for mod in self.pipeline.GetOutputScriptModules():
                self.LoadOutputFilterModule(mod)

            #  load output filters
            for filter in self.pipeline.GetOutputFilters():
                self.lst_output_script.insert('end', filter)
            messagebox.showinfo('',
                        'Pipeline Loaded.')

            self._fill_prompts_combobox()



    def SavePipeline(self):
        if not self.pipeline_name.get().strip():
            messagebox.showwarning('', 'You must set a name for the pipeline.')
            return
        if not self.output_pipeline_filename.get():
            filename = self.SaveAsPipeline()
            if not filename:
                return
            self.output_pipeline_filename.set(filename)
        self.pipeline.SavePipeline(self.output_pipeline_filename.get())
        messagebox.showinfo('', 'Experimental pipeline saved.')

    def AddPipelineItemToPipeline(self):
        if not self.pipeline_item_name.get().strip():
            messagebox.showinfo('',
                                'You must set a name for this pipeline item.')
            return

        if self.pipeline_item_name.get() in list(self.cmb_input_prompt_input_value['values']):
            messagebox.showinfo('',
                                'Pipeline Name already in used, please set a new one.')
            return

        if len(list([v for v in self.cmb_input_prompt_key_name['values'] if v])) > 0:

            messagebox.showerror('',
                                 'You must set each prompt key name before adding this item to the pipeline.')
            return

        #  create pipeline item
        item_input = self.__create_input_item()
        item = PipelineItem()
        item.SetName(self.pipeline_item_name.get())
        item.SetPrompt(self.selected_prompt.get())
        item.SetInput(input_item=item_input)

        self.pipeline.AddItem(deepcopy(item))
        #  add item to pipeline tree view
        self.pipeline_tree.insert('',
                                  'end',
                                  text=str(self.pipeline.total_steps()),
                                  values=item.GetTreeItem())
        #  add item name to input list
        self._add_input_type_item(item.Name())
        self.pipeline_item_name.set('')
        #  fix combo  experiment item
        self.ClearInputKeyValuePairs()
        self._clear_item_output_filter()

    def __create_input_item(self):
        item_input = {'apply-item-filters': [i for i in self.lst_item_filters.get(0, 'end')], #self.chk_apply_filter_to_pipeline_item_output.get(),
                      'input-bindings': list(),
                      }
        for child in self.input_key_pairs.get_children():
            child = self.input_key_pairs.item(child)['values']
            item_input['input-bindings'].append({'prompt-key': child[0],
                                                 'key-value': child[1]})
        return deepcopy(item_input)

######
######
#####


def AskGPT3(prompt,
            gpt_settings):

    # def _extract_answers(answers):
    #     return [ans['text'] for ans in answers['choices']]
    #
    # print('GPT3 Dev Mode')
    # return "this is a list\n of results\n \n   A:   \n -aaaa\n -'b'"
    try:
        response = openai.Completion.create(
                            engine=gpt_settings[PromptDesignerDataset.ENGINE],
                            prompt=prompt,
                            max_tokens=int(gpt_settings[PromptDesignerDataset.MAX_TOKEN]),
                            stop=gpt_settings[PromptDesignerDataset.STOPWORDS_LIST],
                            temperature=gpt_settings[PromptDesignerDataset.TEMPERATURE],
                            top_p=gpt_settings[PromptDesignerDataset.NUCLEUS],
                            n=gpt_settings[PromptDesignerDataset.N_],
                            presence_penalty=gpt_settings[PromptDesignerDataset.PRESENCE_PENALTY],
                            frequency_penalty=gpt_settings[PromptDesignerDataset.FREQUENCY_PENALTY],
                            )
        return response['choices'][0]['text']

    except Exception as err:
        messagebox.showerror('',
                             'OpenAI GPT-3 Error: {}'.format(str(err))
                             )
        return err

def LunchExperiment():

    app = tk.Tk()
    Experiment(app)
    app.mainloop()

    sys.exit(0)

if __name__ == '__main__':
    LunchExperiment()


import sys
import codecs
from copy import deepcopy
import tkinter as tk
import tkinter.ttk as ttk
import importlib
import importlib.machinery
import importlib.util
import os
import openai
from tkinter import messagebox
from tkinter import filedialog
from tkinter import simpledialog
from tkinter import PhotoImage
from pathlib import Path
from promptdesignerdataset.dataset import PromptDesignerDataset, LoadJsonData, SaveJsonData
from PIL import Image, ImageDraw, ImageTk, ImageFont
from itertools import combinations, product
from pipelinemaker.pipeline import Pipeline, PipelineItem
import pprint

def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    base_path = getattr(sys, '_MEIPASS', os.path.dirname(os.path.abspath(__file__)))
    return os.path.join(base_path, relative_path)

class ExperimentGUI: #(tk.Frame):
    __version__ = '1.0.1'
    PADX = 5
    PADY = 2
    IMG_X = 25
    BUTTON_HEIGHT = IMG_X + 5

    DO_NOTHING = 'Do nothing'
    #
    # FOR_EACH_ELEMENT = 'for-each-element'
    # FOR_EACH_PAIR = 'for-each-pair'
    #
    # OUTPUT_LIST_OUTPUT = 'output-list'
    # OUTPUT_LIST_OUTPUT = 'output-list'
    # OUTPUT_SINGLE_OUTPUT = 'output-single'
    #  apply filter on the results
    ITEM_OUTPUT_FILTER_TYPE_RESULTS = 'filter on results'
    #  apply filter on the item
    ITEM_OUTPUT_FILTER_TYPE_ITEM = 'ad-hoc filter'

    def __init__(self,
                 parent,
                 ):
        # tk.Frame.__init__(self, parent)
        self.parent = parent
        self.parent.geometry("1200x700")
        self.parent.title('Prompt Designer - Experiments Maker')

        self.__init_variables__()
        self.__load_icons()
        self.__load_program_icon()
        self.CreateFrames()
        self.CreateMenu()

        self._fill_prompts_combobox()
        self._fill_pipeline_item_output_filter_list()

    def __load_program_icon(self):
        # return
        self.parent.tk.call('wm', 'iconphoto', self.parent._w, self.images['logo'])


    def __init_variables__(self):
        #  handle prompts
        self.promptdesignerdataset = PromptDesignerDataset()
        # item output filter type
        self.item_output_filter_type = tk.StringVar()

        #  store module imported
        self.imported_output_filters = dict()
        self.imported_item_output_filtes = dict()

        self.pipeline_name = tk.StringVar()
        self.pipeline_name.trace_variable(mode='w', callback=self._pipeline_name_changed)

        #  store input file filename
        self.input_file_filename = None
        #  store input type
        self.input_type = tk.StringVar()
        #  store mnemonic name of the input file
        self.input_file_mnemonic_name = tk.StringVar()

        self.selected_setting_filename = tk.StringVar()
        self._prompt_key_name = tk.StringVar()
        self._prompt_input_value = tk.StringVar()
        #  opened input file content
        self.input_text_content = None

        #  dataset completed path
        self.output_pipeline_filename = tk.StringVar()

        #  selected prompt
        self.selected_prompt = tk.StringVar()

        #  pipiline item name
        self.pipeline_item_name = tk.StringVar()

        # handle radio button selection
        #  keep radio vaule for each element or for each pair
        self.experiment_setting_range_type = tk.StringVar()
        #  keep radio value for output type
        self.experiment_output_type = tk.StringVar()

        self.chk_apply_filter_to_pipeline_output = tk.BooleanVar()
        self.chk_apply_filter_to_pipeline_output.set(False)

        self.pipeline_item_output_filter_name = tk.StringVar()
        self.pipeline_output_filter_name = tk.StringVar()

        #  GPT-3 Settings
        self.gpt3settings = dict()

        #  Pipeline
        self.pipeline = Pipeline()

        #  checked status for re-run the entire pipeline or only the missing steps
        self._force_rerun_entire_pipeline = tk.BooleanVar()
        self._force_rerun_entire_pipeline.set(True)

    def __load_icons(self):
        # icon_folder = Path(__file__).absolute()
        # icon_folder = icon_folder.parent
        icon_folder = Path(resource_path('icons'))
        # print(os.path.exists(icon_folder))
        # print(os.path.exists(icon_folder.joinpath('load_file.png')))

        self.images = {'load': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('load_file.png')).resize((self.IMG_X, self.IMG_X))),
                       'save': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('save_file.png')).resize((self.IMG_X, self.IMG_X))),
                       'show': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('show.png')).resize((self.IMG_X, self.IMG_X))),
                       'clear': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('clear.png')).resize((self.IMG_X, self.IMG_X))),

                       'pipeline': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('pipeline.png')).resize((self.IMG_X, self.IMG_X))),
                       'experiment': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('experiment.png')).resize((self.IMG_X, self.IMG_X))),
                       'add': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('add.png')).resize((self.IMG_X, self.IMG_X))),
                       'checked': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('checked.png')).resize((self.IMG_X, self.IMG_X))),

                        'logo': ImageTk.PhotoImage(
                            Image.open(icon_folder.joinpath('logo.png')).resize((self.IMG_X, self.IMG_X))),
                       }
        #
        # self.images = {'load': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('load_file.png')).resize((self.IMG_X, self.IMG_X))),
        #                'save': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('save_file.png')).resize((self.IMG_X, self.IMG_X))),
        #                'show': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('show.png')).resize((self.IMG_X, self.IMG_X))),
        #                'clear': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('clear.png')).resize((self.IMG_X, self.IMG_X))),
        #
        #                'pipeline': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('pipeline.png')).resize((self.IMG_X, self.IMG_X))),
        #                'experiment': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('experiment.png')).resize((self.IMG_X, self.IMG_X))),
        #                'add': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('add.png')).resize((self.IMG_X, self.IMG_X))),
        #                'checked': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('checked.png')).resize((self.IMG_X, self.IMG_X))),
        #
        #                 'logo': ImageTk.PhotoImage(
        #                     Image.open(icon_folder.joinpath('logo.png')).resize((self.IMG_X, self.IMG_X))),
        #                }

    def CreateMenu(self):
        self.menubar = tk.Menu(self.parent)

        self.filemenu = tk.Menu(self.menubar, tearoff=0)
        self.filemenu.add_command(label="Import prompts from prompt dataset", command=self.LoadPromptDesignerDataset)

        self.filemenu.add_separator()
        self.menubar.add_cascade(label="File", menu=self.filemenu)

        gptsettings = tk.Menu(self.menubar, tearoff=0)
        gptsettings.add_command(label="Load GPT-3 settings", command=self._load_gptsettings)
        gptsettings.add_command(label="Show GPT-3 settings", command=self._show_gpt3_settings)
        self.filemenu.add_cascade(label="GPT-3 Settings", menu=gptsettings)
        self.filemenu.add_separator()
        self.filemenu.add_command(label="Exit", command=self.parent.destroy)

        pipeline = tk.Menu(self.menubar, tearoff=0)
        pipeline.add_command(label="New", command=self.NewPipeline)
        pipeline.add_command(label="Load", command=self.LoadPipeline)
        pipeline.add_command(label="Save", command=self.SavePipeline)
        pipeline.add_command(label="Save As", command=self.SaveAsPipeline)
        pipeline.add_separator()
        pipeline.add_command(label="Clear", command=self.ClearPipeline)
        pipeline.add_separator()
        pipeline.add_command(label='Show item', command=self.ShowPipelineItem)
        self.menubar.add_cascade(label="Pipeline", menu=pipeline)

        results = tk.Menu(self.menubar, tearoff=0)
        results.add_command(label="Show results", command=self.ShowResults)
        results.add_command(label="Export results", command=self.ExportResultsGUI)
        self.menubar.add_cascade(label="Results", menu=results)

        filters = tk.Menu(self.menubar, tearoff=0)
        filters.add_command(label="Import item-filter module", command=self.LoadItemOutputModuleGUI)
        filters.add_command(label="Import output script module", command=self.LoadOutputFilterModuleGUI)
        self.menubar.add_cascade(label="Filters and Scripts", menu=filters)

        self.lunch = tk.Menu(self.menubar, tearoff=0)
        self.lunch.add_command(label="Lunch Experiment", command=self.LunchExperimentalPipeline)
        self.lunch.add_command(label="Apply output script", command=self.LunchOutputScript)
        self.lunch.add_separator()
        self.lunch.add_checkbutton(label="force re-run the entire pipeline", onvalue=1, offvalue=0, variable=self._force_rerun_entire_pipeline)
        self.lunch.add_separator()
        self.lunch.add_command(label='run only selected item', command=self.LunchExperimentalPipelineSelectedItem)

        self.menubar.add_cascade(label="Lunch Experiment", menu=self.lunch)

        self.parent.config(menu=self.menubar)
    #
    # def _force_rerun_entire_pipeline_change_status(self):
    #     if self._force_rerun_entire_pipeline.get():
    #         #  set status
    #         self._force_rerun_entire_pipeline.set(False)
    #         #  set icon on menu
    #         self.lunch.entryconfig(3, {'image': None})
    #     else:
    #         #  set status
    #         self._force_rerun_entire_pipeline.set(True)
    #         #  set icon on menu
    #         self.lunch.entryconfig(3, {'image': self.images['checked']})

    def _add_input_type_item(self, input_item):
        if type(self.cmb_input_prompt_input_value['values']) == str:
            self.cmb_input_prompt_input_value['values'] = [input_item]
        else:
            vals = list(self.cmb_input_prompt_input_value['values'])
            vals.append(input_item)
            vals = list(set(vals))
            self.cmb_input_prompt_input_value['values'] = vals

    def CreateFrames(self):
        #   MAIN TOP FRAME
        self.frm_maintop = tk.Frame(self.parent)
        self.frm_maintop.pack(fill='x',
                              # expand=True,
                              side='top')
        self.__create_frame_maintop()

        #   MAIN PIPELINE FRAME
        self.frm_main_pipeline = tk.Frame(self.parent)
        self.frm_main_pipeline.pack(fill='both',
                               expand=True,
                               side='top')
        self.__create_frame_mainpipeline()

    def _clear_output_script(self, *event):
        self.lst_output_script.delete(0, 'end')

    def _add_item_output_filter(self, *event):
        if self.pipeline_item_output_filter_name.get():
            filter_item = tuple([self.pipeline_item_output_filter_name.get(), self.item_output_filter_type.get()])
            self.lst_item_filters.insert('end', filter_item)

    def _fill_pipeline_item_output_filter_list(self):
        self.cmb_pipeline_item_output_filter_name['values'] = list()

    def __create_frame_maintop(self):
        self.__frm_maintop_left = tk.Frame(self.frm_maintop,
                                           width=200,
                                           )
        self.__frm_maintop_left.pack(side='left', fill='both')
        frm_pipeline_name = tk.Frame(self.__frm_maintop_left)
        frm_pipeline_name.pack(side='top', fill='x')
        tk.Label(frm_pipeline_name,
                 text='Pipeline name:',
                 anchor='w',
                 ).pack(side='left')
        tk.Entry(frm_pipeline_name,
                 textvariable=self.pipeline_name,
                 justify='center',
                 ).pack(side='left', fill='x')
        self.__create_frame_maintop_inputfile()
        # self.__create_frame_maintop_gptsettings()
        self.__create_frame_maintop_output()

        self.__frm_maintop_experimentitem = tk.LabelFrame(self.frm_maintop,
                                                          text='Experimental Item',
                                                          labelanchor='n')
        self.__frm_maintop_experimentitem.pack(side='left',
                                               fill='both',
                                               expand=True)
        self.__create_frame_experimentitem()

    def _clear_input_key_pairs(self):
        for item in self.input_key_pairs.get_children():
            self.input_key_pairs.delete(item)

    def _clear_input_values(self):
        self.cmb_input_prompt_input_value['values'] = []
        self.cmb_input_prompt_input_value.set('')

        for menomnic_input_name in self.pipeline.GetInputFilesMnemonicNames():
            self._add_input_type_item(menomnic_input_name)

    def _add_inputfile_into_filetreelist(self,
                                         filename_name,
                                         mnemonic_name,
                                         inputtype):

        self.input_files_tree.insert('',
                                     'end',
                                     text=mnemonic_name,
                                     values=([mnemonic_name, inputtype]))
        self._add_input_type_item(mnemonic_name)

    def __create_frame_maintop_inputfile(self):
        frm_input = tk.LabelFrame(self.__frm_maintop_left,
                                      text='Input Files',
                                      labelanchor='n')
        frm_input.pack(side='top', fill='x', expand=True)
        frm_inputbuttons = tk.Frame(frm_input)
        frm_inputbuttons.pack(side='top',
                              fill='x')
        tk.Button(frm_inputbuttons,
                  text='', #  Select Input File',
                  anchor='w',
                  height=self.BUTTON_HEIGHT,
                  width=self.BUTTON_HEIGHT,
                  image=self.images['load'],
                  compound='center',
                  relief='raised',
                  command=self.SelectInputFile,
                  ).pack(side='left')

        tk.Button(frm_inputbuttons,
                  text='', #Show text',
                  anchor='w',
                  image=self.images['show'],
                  compound='center',
                  height=self.BUTTON_HEIGHT,
                  width=self.BUTTON_HEIGHT,
                  command=self.InputFilePreview,
                  ).pack(side='left')

        tk.Label(frm_inputbuttons, #frm_input_selectedfile,
                 anchor='w',
                 # relief='raised',
                 text='Mnemonic Name',
                 ).pack(side='left')
        tk.Entry(frm_inputbuttons, #frm_input_selectedfile,
                 justify='left',
                 width=15,
                 relief='raised',
                 textvariable=self.input_file_mnemonic_name,
                 ).pack(side='left', fill='x')

        frm_input_type = tk.Frame(frm_input)
        frm_input_type.pack(side='top', fill='x')
        tk.Radiobutton(frm_input_type,
                       text='Entire file',
                       anchor='w',
                       value=Pipeline.INPUT_TYPE_ENTIRE,
                       variable=self.input_type,
                       ).pack(side='left')
        tk.Radiobutton(frm_input_type,
                       text='File is a list',
                       anchor='w',
                       value=Pipeline.INPUT_TYPE_LIST,
                       variable=self.input_type,
                       ).pack(side='left')
        self.input_type.set(Pipeline.INPUT_TYPE_ENTIRE)
        tk.Button(frm_input_type,
                  image=self.images['add'],
                  # text='  Add input file',
                  compound='left',
                  anchor='w',
                  command=self.AddInputFile,
                  ).pack(side='left')
        tk.Button(frm_input_type,
                  image=self.images['clear'],
                  # text='  Clear input file list',
                  compound='left',
                  anchor='w',
                  command=self.ClearInputFiles,
                  ).pack(side='left')

        frm_input_files = tk.Frame(frm_input)
        frm_input_files.pack(side='left', fill='x')

        input_file_columns = tuple(['menomonicname', 'inputtype'])
        self.input_files_tree = ttk.Treeview(frm_input_files,
                                             height=3,
                                             columns=input_file_columns)
        self.input_files_tree.pack(side='top',
                                   fill='both')
        self.input_files_tree.heading('#0', text='text')
        self.input_files_tree.heading('#1', text='mnemonic name')
        self.input_files_tree.heading('#2', text='input type')
        # Specify attributes of the columns (We want to stretch it!)
        self.input_files_tree.column('#0', minwidth=0, width=0, anchor='w',stretch=tk.NO)
        self.input_files_tree.column('#1', anchor='w', stretch=tk.YES)
        self.input_files_tree.column('#2', anchor='w', stretch=tk.YES)

    def _load_gptsettings(self):
        filename = Path(filedialog.askopenfilename(initialdir = os.getcwd(),
                                                   title = "Please select a file",
                                                   filetypes = (('json file', '*.json'),),
                                                   defaultextension='.json'
                                                   )
                        )
        if filename.name:
            #  read file
            self.selected_setting_filename.set(filename.name)
            self.gpt3settings = LoadJsonData(filename.absolute())
            #  add file ref
            self.gpt3settings['file'] = str(filename.absolute())

            try:
                #  set openai.key
                openai.api_key = self.gpt3settings["api-key"]
            except KeyError:
                messagebox.showerror('',
                                     'no valid API-key found!. Setting NOT loaded.')
                return

            self.pipeline.AddGPT3Settings(self.gpt3settings,)
            if self.gpt3settings:
                messagebox.showinfo('',
                                    'GPT-3 Settings loaded.')
            else:
                messagebox.showerror('',
                                     'Setting file corrupted.')

    def SelectInputFile(self):
        filename = Path(filedialog.askopenfilename(initialdir=os.getcwd(),
                                                   title="Please select a file",
                                                   filetypes=(('all files', '*.*'),
                                                              ('text file', '*.txt'),
                                                              ),
                                                   defaultextension='.txt'
                                                   )
                        )
        if filename.name:
            self.input_file_filename = str(filename.absolute())
            self.input_file_mnemonic_name.set(filename.name)

    def __create_frame_maintop_output(self):
        frm_output = tk.LabelFrame(self.__frm_maintop_left,
                                         text='Output',
                                         labelanchor='n')
        frm_output.pack(side='top',
                        fill='both',
                        expand=True)

        frm_output_script = tk.Frame(frm_output)
        frm_output_script.pack(side='top',
                               fill='x',
                               expand=True)

        tk.Button(frm_output_script,
                  # text='import module',
                  image=self.images['load'],
                  compound='center',
                  anchor='w',
                  height=self.BUTTON_HEIGHT,
                  command=self.LoadOutputFilterModuleGUI,
                  ).pack(side='left')
        self.cmb_pipeline_output_filter_name = ttk.Combobox(frm_output_script,
                                                            textvariable=self.pipeline_output_filter_name,
                                                            )
        self.cmb_pipeline_output_filter_name.pack(side='left',
                                                  fill='x',
                                                  expand=True)
        tk.Button(frm_output_script,
                  # text=' Add filter',
                  image=self.images['add'],
                  compound='center',
                  anchor='w',
                  height=self.BUTTON_HEIGHT,
                  command=self._add_output_script,
                  ).pack(side='left')
        tk.Button(frm_output_script,
                  # text=' Clear filters',
                  image=self.images['clear'],
                  compound='center',
                  anchor='w',
                  height=self.BUTTON_HEIGHT,
                  command=self._clear_output_script).pack(side='left')
        self.lst_output_script = tk.Listbox(frm_output,
                                            height=5, )
        self.lst_output_script.pack(side='top', fill='x')

    def _add_imported_output_filters_cmb(self, script_name):
        if type(self.cmb_pipeline_output_filter_name['values']) == str:
            self.cmb_pipeline_output_filter_name['values'] = [script_name]
        else:
            vals = list(self.cmb_pipeline_output_filter_name['values'])
            vals.append(script_name)
            vals = list(set(vals))
            self.cmb_pipeline_output_filter_name['values'] = vals
        self.pipeline_output_filter_name.set(self.cmb_pipeline_output_filter_name['values'][0])

    def _clear_item_output_filter(self, *event):
        self.lst_item_filters.delete(0, 'end')

    def _load_inputfile_content(self,
                                filename):
        if self.input_type.get() == Pipeline.INPUT_TYPE_ENTIRE:
            return codecs.open(filename, 'r', 'utf-8').read()
        else:
            # it is a list
            return [line.strip() for line in codecs.open(filename, 'r', 'utf-8').readlines() if line.strip()]

    def __create_frame_experimentitem(self):
        frm_experiment_item = tk.Frame(self.__frm_maintop_experimentitem)
        frm_experiment_item.pack(side='left',
                                 fill='both',
                                 expand=True)
        frm_explabel = tk.Frame(frm_experiment_item)
        frm_explabel.pack(side='top',
                          fill='x')
        tk.Label(frm_explabel,
                 text='Pipeline item name:'
                 ).pack(side='left')
        tk.Entry(frm_explabel,
                 textvariable=self.pipeline_item_name,
                 ).pack(side='left', fill='x', expand=True)


        frm_prompt = tk.Frame(frm_experiment_item)
        frm_prompt.pack(side='top', fill='x', expand=True)

        tk.Label(frm_prompt,
                 text='Prompt:',
                 anchor='w'
                 ).pack(side='left')
        self.cmb_prompt = ttk.Combobox(frm_prompt,
                                       text='Prompt List',
                                       textvariable=self.selected_prompt)
        self.cmb_prompt.pack(side='left',
                             fill='x',
                             expand=True)
        self.cmb_prompt.bind("<<ComboboxSelected>>", self._cmb_prompt_changed)
        self.cmb_prompt['state'] = 'readonly'

        tk.Button(frm_prompt,
                  # text='Show prompt',
                  image=self.images['show'],
                  compound='left',
                  # height=self.BUTTON_HEIGHT,
                  width=self.BUTTON_HEIGHT,
                  command=self._show_prompt,
                  ).pack(side='right')

        frm_key_val_filter = tk.Frame(frm_experiment_item)
        frm_key_val_filter.pack(side='top',
                  fill='x',
                  expand=True)

        frm_key_pair = tk.LabelFrame(frm_key_val_filter,
                                     text='Key-Value Bindings',
                                     labelanchor='n')
        frm_key_pair.pack(side='left',
                  fill='x',
                  expand=True)

        frm_expkey = tk.Frame(frm_key_pair)
        frm_expkey.pack(side='top', fill='x')
        tk.Label(frm_expkey,
                 text='prompt key',
                 ).pack(side='left', fill='x')
        self.cmb_input_prompt_key_name = ttk.Combobox(frm_expkey,
                                                      textvariable=self._prompt_key_name)
        self.cmb_input_prompt_key_name.pack(side='left', fill='x', expand=True)
        self.cmb_input_prompt_key_name['state'] = 'readonly'

        frm_expval = tk.Frame(frm_key_pair)
        frm_expval.pack(side='top', fill='x')
        tk.Label(frm_expval,
                 text='input value',
                 ).pack(side='left', fill='x')
        self.cmb_input_prompt_input_value = ttk.Combobox(frm_expval,
                                                         textvariable=self._prompt_input_value)
        self.cmb_input_prompt_input_value.pack(side='left', fill='x', expand=True)
        self.cmb_input_prompt_input_value['state'] = 'readonly'

        frm_buttons = tk.Frame(frm_key_pair)
        frm_buttons.pack(side='top', fill='x')

        tk.Button(frm_buttons,
                  text='Add key-value pair',
                  image=self.images['add'],
                  compound='left',
                  height=self.BUTTON_HEIGHT,
                  command=self._add_input_key_pair,
                  ).pack(side='left', fill='x', expand=True)

        tk.Button(frm_buttons,
                  text='delete pairs',
                  image=self.images['clear'],
                  compound='left',
                  height=self.BUTTON_HEIGHT,
                  command=self.ClearInputKeyValuePairs,
                  ).pack(side='left', fill='x', expand=True)


        frm_tree = tk.Frame(frm_key_pair)
        frm_tree.pack(side='top',
                      fill='both',
                      expand=True)
        input_key_pairs_columns = (['Prompt Key', 'type'])
        self.input_key_pairs = ttk.Treeview(frm_tree,
                                            columns=input_key_pairs_columns,
                                            height=5,
                                            )
        self.input_key_pairs.pack(side='left',
                                  fill='both',
                                  expand=True)
        # Set the heading (Attribute Names)
        self.input_key_pairs.heading('#0', text='text')
        self.input_key_pairs.heading('#1', text='Prompt Key')
        self.input_key_pairs.heading('#2', text='Input')
        # Specify attributes of the columns (We want to stretch it!)
        self.input_key_pairs.column('#0', minwidth=0, width=0, anchor='center',stretch=tk.NO)
        self.input_key_pairs.column('#1', anchor='center', stretch=tk.YES)
        self.input_key_pairs.column('#2', anchor='center', stretch=tk.YES)

        frm_filter = tk.LabelFrame(frm_key_val_filter,
                                   text='Item Output Filters',
                                   labelanchor='n')
        frm_filter.pack(side='right',
                  fill='x',
                  expand=True)

        frm_filter_buttons = tk.Frame(frm_filter)
        frm_filter_buttons.pack(side='top',
                        fill='x',
                        expand=True)
        tk.Button(frm_filter_buttons,
                  # text='  load filter module',
                  image=self.images['load'],
                  compound='left',
                  anchor='w',
                  command=self.LoadItemOutputModuleGUI,
                  ).pack(side='left', fill='x')
        self.cmb_pipeline_item_output_filter_name = ttk.Combobox(frm_filter_buttons,
                                                                 # text='select filter',
                                                                 textvariable=self.pipeline_item_output_filter_name,
                                                                 justify='left',
                                                                 )
        self.cmb_pipeline_item_output_filter_name.pack(side='left',
                                                       fill='x',
                                                       expand=True)
        self.cmb_pipeline_item_output_filter_name['state'] = 'readonly'

        frm_filter_type = tk.Frame(frm_filter)
        frm_filter_type.pack(side='top', fill='both')

        tk.Radiobutton(frm_filter_type,
                       text='on result',
                       variable=self.item_output_filter_type,
                       value=self.ITEM_OUTPUT_FILTER_TYPE_RESULTS,
                       anchor='w',
                       ).pack(side='left', fill='x')
        tk.Radiobutton(frm_filter_type,
                       text='ah-hoc on item',
                       variable=self.item_output_filter_type,
                       value=self.ITEM_OUTPUT_FILTER_TYPE_ITEM,
                       anchor='w',
                       ).pack(side='left', fill='x')
        #  set default value
        self.item_output_filter_type.set(self.ITEM_OUTPUT_FILTER_TYPE_RESULTS)

        frm_btns = tk.Frame(frm_filter)
        frm_btns.pack(side='top', fill='x')
        tk.Button(frm_btns,
                  text='  Add filter',
                  image=self.images['add'],
                  compound='left',
                  anchor='w',
                  command=self._add_item_output_filter,
                  ).pack(side='left', fill='x')
        tk.Button(frm_btns,
                  text='  Clear filters',
                  image=self.images['clear'],
                  compound='left',
                  anchor='w',
                  command=self._clear_item_output_filter,
                  ).pack(side='left', fill='x')

        self.lst_item_filters = tk.Listbox(frm_filter,
                                           # height=5,
                                           )
        self.lst_item_filters.pack(side='top', fill='both')

        tk.Button(frm_experiment_item,
                  text='   Add experimental item to pipeline',
                  anchor='center',
                  command=self.AddPipelineItemToPipeline,
                  image=self.images['add'],
                  compound='left',
                  height=self.BUTTON_HEIGHT,
                  ).pack(side='top',
                         fill='x',
                         expand=True)

    def _add_imported_item_output_filters_cmb(self, filter_name):
        if type(self.cmb_pipeline_item_output_filter_name['values']) == str:
            self.cmb_pipeline_item_output_filter_name['values'] = [filter_name]
        else:
            vals = list(self.cmb_pipeline_item_output_filter_name['values'])
            vals.append(filter_name)
            vals = list(set(vals))
            self.cmb_pipeline_item_output_filter_name['values'] = vals

        self.pipeline_item_output_filter_name.set(self.cmb_pipeline_item_output_filter_name['values'][0])

    def _fill_prompts_combobox(self):
        vals = list(self.pipeline.GetPromptNames())
        vals.insert(0, self.DO_NOTHING)
        self.cmb_prompt['values'] = vals

        if len(vals) > 1:
            self.selected_prompt.set(self.cmb_prompt['values'][1])
        else:
            self.selected_prompt.set(self.cmb_prompt['values'][0])
        self._cmb_prompt_changed()

    def _add_output_script(self, *event):
        if self.pipeline_output_filter_name.get():
            self.lst_output_script.insert('end', self.pipeline_output_filter_name.get())
            self.pipeline.AddOutputFilter(self.pipeline_output_filter_name.get())

    def _clear_imported_output_filters(self):
        self.cmb_pipeline_output_filter_name['values'] = []
        self.cmb_pipeline_output_filter_name.set('')
        self.pipeline.ClearOutputFilters()

    def _clear_imported_item_output_filters(self):
        self.cmb_pipeline_item_output_filter_name['values'] = []
        self.cmb_pipeline_item_output_filter_name.set('')
        # self.pipeline.ClearOutputFilters()

    def _add_input_key_pair(self, *event):
        if not self._prompt_key_name.get().strip() or not self._prompt_input_value.get().strip():
            return
        self.input_key_pairs.insert('',
                                  'end',
                                  text=self._prompt_key_name.get(),
                                  values=tuple([self._prompt_key_name.get(),
                                                self._prompt_input_value.get()]))
        #  remove used key from cmb list
        vals = list(self.cmb_input_prompt_key_name['values'])
        vals.remove(self._prompt_key_name.get())
        self.cmb_input_prompt_key_name['values'] = vals
        self._prompt_key_name.set('')

    def __create_frame_mainpipeline(self):
        #  macro container
        frm_pipeline = tk.LabelFrame(self.frm_main_pipeline,
                                     text='Pipeline',
                                     labelanchor='n')
        frm_pipeline.pack(side='top',
                         fill='both',
                         expand=True)

        pipeline_tree_columns = ('Experimantal Item', 'input',  'Prompt Name', 'Filter Name')
        self.pipeline_tree = ttk.Treeview(frm_pipeline,
                                          columns=pipeline_tree_columns
                                        )
        self.pipeline_tree.pack(side='top',
                                fill='both',
                                expand=True)
        self.pipeline_tree. bind('<Double-Button-1>', self.RemovePipelineItem)
        # Set the heading (Attribute Names)
        self.pipeline_tree.heading('#0', text='Step ID')
        self.pipeline_tree.heading('#1', text='Experimental Item')
        self.pipeline_tree.heading('#2', text='Prompt Name')
        self.pipeline_tree.heading('#3', text='Input')
        # self.pipeline_tree.heading('#4', text='Filter on Output')
        self.pipeline_tree.heading('#4', text='Filter list')

        # Specify attributes of the columns (We want to stretch it!)
        self.pipeline_tree.column('#0',
                                  anchor='center',
                                  width=50,
                                  minwidth=50,
                                  stretch=tk.NO)
        for col_n in range(1, len(pipeline_tree_columns)):
            col_id = '#{}'.format(col_n)
            self.pipeline_tree.column(col_id, anchor='center',
                                  minwidth=50,
                                  stretch=tk.YES)



from setuptools import find_packages, setup
import os

def read(fname):
    """
    Utility function to read the README file. Used for the long_description.
    :param fname:
    :return:
    """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

_REQUIRED = ['Pillow', 'openai',  'promptdesignerdataset']
_EXTRAS = []
setup(
    name='pipelinemaker',
    version='0.0.2a5',
    packages=find_packages(),
    install_requires=_REQUIRED,
    # extras_require=_EXTRAS,
    python_requires='!=2.7, >=3.5.*',
    url='https://pdi.fbk.eu/process-extraction-from-text-overview/',
    license='MIT',
    classifiers=[
            "Development Status :: 3 - Alpha",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",

            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: End Users/Desktop",
            "Intended Audience :: Information Technology",

            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",

            "Topic :: Scientific/Engineering",
            "Topic :: Utilities",
        ],
keywords=["huggingface", "PET", "dataset", "benchmark", "process extraction from text",
              "natural language processing", "nlp", "business process management", "bpm"],
    author='Patrizio Bellan',
    include_package_data=True,
    long_description=read("README.rst"),
    long_description_content_type='text/x-rst',
    author_email='patrizio.bellan@gmail.com',
    description='pipeline maker. this tool allows to create custom GPT-3 pipeline easily'
)
